/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

/**
 * \file
 * \ingroup Implementation
 */


#ifndef _NAIVE_DECISIONER_H
#define _NAIVE_DECISIONER_H

#include "policy.hpp"
#include "global_nodestruct.hpp"
#include "storage_backend_serialized.hpp"

#include <map>
#include <vector>

namespace ldp_xml_parser
{
	typedef std::vector<gid_t> VGid;

	/* Class which checks rights in policies retrieved from policy db
	 * \ingroup Implementation
	 */
	class NaivePolicyChecker {
	private:

		/** Policy database */
		ldp_serialized::StorageBackendSerialized m_bus_db;

		/** Parses delivered decision. In case of Decision::CHECK calls cynara.
		 * \param[in] decision Decision from checkers
		 * \param[in] uid User id
		 * \param[in] label User label
		 * \return Returns deny=0, allow=1 or cynara error
		 * \ingroup Implementation
		 */
		DecisionResult parseDecision(const DecisionItem& decision,
						   uid_t uid,
						   const char* label) const;

		/** Checks policy for a given own, send or receive item
		 * \param[in] uid User id
		 * \param[in] gid User group id
		 * \param[in] item Item to check
		 * \return Returns deny=0, allow=1 or cynara error
		 * \ingroup Implementation
		 */
		template<typename T>
		DecisionItem checkItem(uid_t uid,
					   gid_t gid,
					   const T& item);

		/** Checks policy for a given access item
		 * \param[in] item Item to check
		 * \return Returns deny=0, allow=1 or cynara error
		 * \ingroup Implementation
		 */
		DecisionItem checkItemAccess(const MatchItemAccess &item);

		template<typename T>
		DecisionItem checkGroupPolicies(uid_t uid,
						gid_t gid,
						const T& item);

		/* group maps management */
		std::map<uid_t, VGid> mapGroups[MatchItemTypes::count];
		/* A mutex for mapGroups */
		pthread_mutex_t mutexGroup = PTHREAD_MUTEX_INITIALIZER;

		template <typename T>
		VGid &getMapGroup(uid_t uid);

		void updateSupplementaryGroups(const VGid &groups, uid_t uid, gid_t gid);
		void updateSupplementaryGroupsOwn(const VGid &groups, uid_t uid, gid_t gid);

	public:
		/** Retrieves policy db
		 * \return Returns reference to the policy db
		 */
		const decltype(m_bus_db) &getPolicyDb() const { return m_bus_db; }

		/** Clears all db data, useful for reloading configuration
		 * during testing.
		 */
		bool initDb(const char *config_name, const char *serialized_filename = nullptr);

		void updateGroupDb(uid_t uid, gid_t gid);

		template <typename T>
		const VGid *getGroups(uid_t uid, gid_t gid);

		/** Prints to stderr the structures and the amount of their memory */
		void printContent();

		/** Checks access/open policy for given item
		 * \param[in] uid User id
		 * \param[in] gid User group id
		 * \return Returns deny=0, allow=1 or cynara error
		 * \ingroup Implementation
		 */
		DecisionResult check(uid_t bus_owner,
				   uid_t uid,
				   gid_t gid,
				   const char* const label);

		/** Checks send/receive/ownership policy for given item
		 * \param[in] uid User id
		 * \param[in] gid User group id
		 * \param[in] label User label
		 * \param[in] matchItem MatchItem to check
		 * \return Returns deny=0, allow=1 or cynara error
		 * \ingroup Implementation
		 */
		template <typename T>
		DecisionResult check(uid_t uid,
				   gid_t gid,
				   const char* const label,
				   const T &matchItem);
	};
}

DCL_NODESTRUCT_GLOBAL(ldp_xml_parser::NaivePolicyChecker, policy_checker_system)
DCL_NODESTRUCT_GLOBAL(ldp_xml_parser::NaivePolicyChecker, policy_checker_session)

#endif
