/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
#include "policy.hpp"
#include "print_content.hpp"
#include "serialized_convert.hpp"
#include <boost/algorithm/string/predicate.hpp>
#include <ostream>

/************ PRINTING CONTENT ********************/
std::ostream &print_content_item_own(std::ostream &stream, const std::string &name, bool is_prefix) {
	return stream << "ItemOwn: service(" << (name.empty() ? "*" : name) << "), pref(" << is_prefix << ")";
}

namespace {
const char* message_access[] = {"USER", "GROUP", "ALL_USERS", "ALL_GROUPS"};
inline const char* __access_type_to_str(ldp_xml_parser::BusAccessType type) {
	return message_access[static_cast<std::size_t>(type)];
}
}

namespace print_content {
static bool xml_format = false;
void use_xml_format(const bool xml) {
	xml_format = xml;
}
}

template <typename T>
std::ostream &print_val(std::ostream &stream, const boost::string_ref &name, const T &val) {
	return stream << name << "(" << val << ")";
}

template <typename T>
std::ostream &print_next_val(std::ostream &stream, const boost::string_ref &name, const T &val) {
	stream << ", ";
	return print_val(stream, name, val);
}

std::ostream &print_content_item_access(std::ostream &stream,
										ldp_xml_parser::BusAccessType type,
										uid_t uid,
										gid_t gid,
										const ldp_xml_parser::DecisionItem &decisionItem) {
	stream << "ItemAccess: ";
	print_val(stream, "type", __access_type_to_str(type));
	print_next_val(stream, "uid", uid);
	print_next_val(stream, "gid", gid);
	return print_next_val(stream, "decision", decisionItem);
}

namespace {
static const char* message_type[] = { "ANY", "METHOD_CALL", "METHOD_RETURN", "ERROR", "SIGNAL"};
static inline const char* __message_type_to_str(ldp_xml_parser::MessageType type) {
	return message_type[static_cast<std::size_t>(type)];
}
}

std::ostream &print_content_item_sr(std::ostream &stream,
									const boost::string_ref &item_type,
									const boost::string_ref &name,
									const boost::string_ref &interface,
									const boost::string_ref &member,
									const boost::string_ref &path,
									ldp_xml_parser::MessageType type,
									const ldp_xml_parser::DecisionItem &decisionItem,
									const bool is_prefix)
{
	stream << item_type << ": ";
	if (is_prefix)
		print_val(stream, "name_prefix", name);
	else
		print_val(stream, "name", name);
	print_next_val(stream, "inter", interface);
	print_next_val(stream, "member", member);
	print_next_val(stream, "path", path);
	print_next_val(stream, "type", __message_type_to_str(type));
	return print_next_val(stream, "decision", decisionItem);
}

std::ostream &print_content_item_sr_xml(std::ostream &stream,
									const boost::string_ref &item_type,
									const boost::string_ref &name,
									const boost::string_ref &interface,
									const boost::string_ref &member,
									const boost::string_ref &path,
									ldp_xml_parser::MessageType type,
									const ldp_xml_parser::DecisionItem &decisionItem,
									const bool is_prefix)
{
	const char *strDecision[] = {"any", "allow", "deny", "check"};
	const char *sr;
	std::string type_str;

	if (item_type == "ItemSend")
		sr = "send";
	else
		sr = "receive";

	stream << "<" << strDecision[static_cast<std::size_t>(decisionItem.getDecision())] << " ";
	if (is_prefix)
		stream << sr << "_destination_prefix=\"" << name << "\" ";
	else if (!name.empty())
		stream << sr << "_" << (item_type == "ItemSend" ? "destination" : "sender") << "=\"" << name << "\" ";

	if (!path.empty())
		stream << sr << "_" << "path=\"" << path << "\" ";
	if (!interface.empty())
		stream << sr << "_" << "interface=\"" << interface << "\" ";
	if (!member.empty())
		stream << sr << "_" << "member=\"" << member << "\" ";
	if (!decisionItem.getPrivilege().empty())
		stream << sr << "_" << "privilege=\"" << decisionItem.getPrivilege() << "\" ";

	type_str = __message_type_to_str(type);
	std::transform(type_str.begin(), type_str.end(), type_str.begin(), [](unsigned char c){ return std::tolower(c); });

	if (type_str != "any")
		stream << sr << "_" << "type=\"" << type_str << "\" ";

	return stream << "/>";
}

namespace {
static const char* message_decision[] = {"NO_DECISION", "ALLOW", "DENY", "CHECK"};
static inline const char* __decision_to_str(ldp_xml_parser::Decision dec) {
	return message_decision[static_cast<std::size_t>(dec)];
}
}

std::ostream &print_content_decision_item(std::ostream &stream,
										  ldp_xml_parser::Decision decision,
										  const boost::string_ref &privilege) {
	stream << __decision_to_str(decision);
	if (!privilege.empty())
		stream << ":" << privilege;
	return stream;
}

template <typename T>
void printContentItem(std::ostream &stream, const T *item);

template <> void printContentItem(std::ostream &stream, const FB::ItemAccess *item) {
	print_content_item_access(stream, makeBusAccessType(item->type()), item->uid(),
			item->gid(), makeDecisionItem(item->decision()));
}

template <typename T>
void printContentItemSR(std::ostream &stream, const boost::string_ref &item_type, const T *item) {
	auto print_func = (print_content::xml_format ? print_content_item_sr_xml : print_content_item_sr);
	print_func(stream, item_type, item->name()->c_str(), item->interface()->c_str(),
			item->member()->c_str(), item->path()->c_str(), makeMessageType(item->type()),
			makeDecisionItem(item->decision()), item->is_name_prefix());
}

template <> void printContentItem(std::ostream &stream, const FB::ItemSend *item) {
	printContentItemSR(stream, "ItemSend", item);
}

template <> void printContentItem(std::ostream &stream, const FB::ItemReceive *item) {
	printContentItemSR(stream, "ItemReceive", item);
}

template <typename T>
void printContentPolicy(std::ostream &stream, const T *policy) {
	for (const auto *i: *policy->items()) {
		printContentItem(stream, i);
		stream << std::endl;
	}
}

void printDecisionItem(std::ostream &stream, const FB::DecisionItem *item) {
	print_content_decision_item(stream, makeDecision(item->decision()), item->privilege()->c_str());
}

void printTreeLevel(std::ostream &stream, const FB::PolicyOwnNode *node, unsigned indent) {
	for (decltype(indent) i = 0; i < indent; ++i)
		stream << "\t";
	stream << "| " << node->token()->c_str() << " (" << node->children()->size() << ") | ";
	printDecisionItem(stream, node->decision_item());
	stream << " ";
	printDecisionItem(stream, node->prefix_decision_item());
	stream << std::endl;

	for (const auto &i: *node->children())
		printTreeLevel(stream, i, indent+1);
}

template <> void printContentPolicy(std::ostream &stream, const FB::PolicyOwn *policy) {
	printTreeLevel(stream, policy->tree(), 0);
}

template <typename T>
void printContentUserGroup(std::ostream &stream, const T *set) {
	for (const auto *i: *set->user()) {
		stream << "user: " << i->id() << std::endl;
		printContentPolicy(stream, i->policy());
	}

	for (const auto *i: *set->group()) {
		stream << "group: " << i->id() << std::endl;
		printContentPolicy(stream, i->policy());
	}
}

template <> void printContentUserGroup(std::ostream &, const FB::AccessSet *) {}

template <typename T>
void printContentSet(std::ostream &stream, const T *set) {
	stream << "context default" << std::endl;
	printContentPolicy(stream, set->context_default());
	stream << "context mandatory" << std::endl;
	printContentPolicy(stream, set->context_mandatory());

	printContentUserGroup(stream, set);
}

namespace FB {
std::ostream &operator<<(std::ostream &stream, const FB::File &file) {
	printContentSet(stream, file.m_own_set());
	printContentSet(stream, file.m_send_set());
	printContentSet(stream, file.m_receive_set());
	printContentSet(stream, file.m_access_set());
	return stream;
}
}

/********* PRINTING ITEMS PARSED FROM XML *******************/
namespace ldp_xml_parser {
std::ostream &operator<<(std::ostream &stream, const ldp_xml_parser::ItemOwn &item) {
	return print_content_item_own(stream, item.getName(), item.isPrefix());
}

std::ostream &operator<<(std::ostream &stream, const ldp_xml_parser::ItemSend &item) {
	auto print_func = (print_content::xml_format ? print_content_item_sr_xml : print_content_item_sr);
	return print_func(stream, "ItemSend", item.getName(), item.getInterface(), item.getMember(),
			item.getPath(), item.getType(), item.getDecision(), item.isNamePrefix());
}

std::ostream &operator<<(std::ostream &stream, const ldp_xml_parser::ItemReceive &item) {
	auto print_func = (print_content::xml_format ? print_content_item_sr_xml : print_content_item_sr);
	return print_func(stream, "ItemReceive", item.getName(), item.getInterface(), item.getMember(),
			item.getPath(), item.getType(), item.getDecision(), item.isNamePrefix());
}

std::ostream &operator<<(std::ostream &stream, const ldp_xml_parser::ItemAccess &item) {
	return print_content_item_access(stream, item.getType(), item.getUid(), item.getGid(), item.getDecision());
}

std::ostream &operator<<(std::ostream &stream, const ldp_xml_parser::DecisionItem &item) {
	return print_content_decision_item(stream, item.getDecision(), item.getPrivilege());
}
std::ostream &operator<<(std::ostream& stream, const MatchItemOwn &item)
{
	return stream << (item.getName().empty() ? "NULL" : item.getName());
}

std::ostream &operator<<(std::ostream& stream, const ldp_xml_parser::MatchItemSR &item)
{
	stream << ": services(";
	for (const auto &name: item.names)
		stream << name << " ";

	return stream << "), interface(" << item.interface << "), member(" << item.member <<
		"), path(" << item.path << "), type(" << __message_type_to_str(item.type) << ")";
}

std::ostream &operator<<(std::ostream& stream, const ldp_xml_parser::MatchItemSend &item)
{
	stream << "MatchItemSend";
	return stream << static_cast<const MatchItemSR>(item);
}

std::ostream &operator<<(std::ostream& stream, const ldp_xml_parser::MatchItemReceive &item)
{
	stream << "MatchItemReceive";
	return stream << static_cast<const MatchItemSR>(item);
}

std::ostream &operator<<(std::ostream& stream, const ldp_xml_parser::MatchItemAccess &item)
{
	stream << "uid: " << item.getUid() << ", gid: ";
	for (auto gid : item.getGids())
		stream << gid << ", ";
	return stream << std::endl;
}

}
