/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
#pragma once

#include <utility>

namespace transaction_guard {

/* Helping class for automatically releasing acquired resources when transaction
 * during initialization failed at some point.
 */
template <typename FunRelease>
class Guard {
	FunRelease fun;
	bool active;
public:
	explicit Guard(FunRelease f) : fun(f), active(true) {}
	Guard() = delete;
	Guard(const Guard &) = delete;
	Guard &operator=(const Guard &) = delete;

	~Guard() {
		if (active)
			fun();
	}

	void dismiss() {
		active = false;
	}

	Guard(Guard &&g) : fun(std::move(g.fun)), active(g.active) {
		g.dismiss();
	}
};

template <typename FunRelease>
Guard<FunRelease> makeGuard(FunRelease f) {
	return Guard<FunRelease>(std::move(f));
}

} // namespace transaction_guard
