/*
 * Copyright (c) 2015-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
/**
 * \file
 * \ingroup Implementation
 */

#ifndef _TSLOG_HPP
#define _TSLOG_HPP

#define LOG_TAG "LIBDBUSPOLICY"
#include <dlog.h>
#include <iostream>
#include <thread>
#include <pthread.h>
#include <stdlib.h>
#include <string>

namespace tslog
{
	enum class ldp_log_level {
		SILENT,
		ERROR,
		WARNING,
		DEFAULT = WARNING,
		DEBUG,
		VERBOSE
	};
	/** Checks value of environmental variable with given name */
	bool get_log_env(char const *name);

	/** Checks environmental variables and sets global variable defining if logs are enabled */
	void init(ldp_log_level level = ldp_log_level::SILENT);

	/** Checks if logs are enabled */
	bool enabled(ldp_log_level level = ldp_log_level::DEBUG);

	void flush(ldp_log_level level = ldp_log_level::ERROR);

	void logError(const std::string &error);
	void logWarning(const std::string &warning);
	void logDebug(const std::string &debug);
	void logVerbose(const std::string &verbose);

	template <typename ...Args>
	void log_to_stream(std::ostream &stream, const Args &...args) {
		std::initializer_list<int>{((void)(stream << args), 0)...};
	}

	template <typename ...Args>
	void log_debug(const Args &...args) { if (enabled(ldp_log_level::DEBUG)) log_to_stream(std::cout, args...); }

	template <typename ...Args>
	void log_error(const Args &...args) { if (enabled(ldp_log_level::ERROR)) log_to_stream(std::cout, args...); }

	template <typename ...Args>
	void log_warning(const Args &...args) { if (enabled(ldp_log_level::WARNING)) log_to_stream(std::cout, args...); }

	template <typename ...Args>
	void log_verbose(const Args &...args) { if (enabled(ldp_log_level::VERBOSE)) log_to_stream(std::cout, args...); }

	struct print_errno {
		int err;
		explicit print_errno(int e) : err(e) {}
	};

	std::ostream &operator<<(std::ostream &stream, const print_errno &p);

	/* Use this class to ensure that logs from a given function are not interlaced
	 * with logs from another function
	 */
	class LogLock {
		bool locked;
	public:
		explicit LogLock(ldp_log_level level = ldp_log_level::ERROR);
		~LogLock();
	};
}

#endif
