/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
/**
	return a;
 * \file
 * \ingroup Implementation
 */
#include "tslog.hpp"
#include "uid_gid_helpers.hpp"
#include "xml_parser.hpp"
#include <boost/algorithm/string/predicate.hpp>
#include <cstring>
#include <dirent.h>
#include <expat.h>
#include <functional>
#include <libgen.h>
#include <memory>
#include <unistd.h>

using namespace ldp_xml_parser;

// To prevent huge memory allocation, define max.
// adjustable.
#define MAX_CONF_SIZE (1024 * 1024)

std::string expandPath(const std::string& parent_dir, const std::string& path) {
	if (path[0] == '/') {
		return path;
	}
	return parent_dir + "/" + path;
}

std::string getDir(const char *path) {
	std::string p(path);
	return dirname(&p.front());
}

class XmlParser::IncludeItem {
public:
	std::string filename;
	bool ignore_missing{false};
	bool ignore_always{false};
};

#define parseAssert(cond) \
{ \
	assert(cond); \
	if (!(cond)) { \
		throw std::runtime_error("Parsing error"); \
	} \
}

void start_element_handler(void *data, const char *el, const char **attr) {
	XmlParser& parser = *reinterpret_cast<XmlParser*>(data);
	try {
		parser.elementStart(el, attr);
	} catch (...) {
		tslog::logError("Exception occurred while processing start element.");
		parser.setRetCode(-1);
	}
}

void end_element_handler(void *data, const char *el) {
	XmlParser& parser = *reinterpret_cast<XmlParser*>(data);
	try {
		parser.elementEnd(el);
	} catch (...) {
		tslog::logError("Exception occurred while processing end element.");
		parser.setRetCode(-1);
	}
}

void text_handler(void *data, const char *text, int len) {
	XmlParser& parser = *reinterpret_cast<XmlParser*>(data);
	try {
		parser.text(text, len);
	} catch (...) {
		tslog::logError("Exception occurred while processing element text.");
		parser.setRetCode(-1);
	}
}

bool XmlParser::isMainConfFile(const char *filename) {
	return strcmp(main_filename, filename) == 0;
}

void XmlParser::setRetCode(int ret) {
	ret_code = ret;
}

void XmlParser::text(const char *text, int len) {
	current_text.append(text, len);
}

void XmlParser::elementStart(const char *el, const char **attr) {
	current_text = std::string();

	if (std::strcmp(el, "busconfig") == 0) {
		parseAssert(state == IDLE);
		state = BUSCONFIG;
	} else if (std::strcmp(el, "policy") == 0) {
		parseAssert(state == BUSCONFIG);
		state = POLICY;
		parsePolicyTag(attr);
	} else if (std::strcmp(el, "includedir") == 0) {
		parseAssert(state == BUSCONFIG);
		state = INCLUDEDIR;
	} else if (std::strcmp(el, "include") == 0) {
		parseAssert(state == BUSCONFIG);
		state = INCLUDE;
		ignore_missing = false;
		ignore_always = false;
		for (int i = 0; attr[i]; i += 2) {
			if ((std::strcmp(attr[i], "ignore_missing") == 0) && (std::strcmp(attr[i + 1], "yes") == 0)) {
				ignore_missing = true;
			} else if ((std::strcmp(attr[i], "if_selinux_enabled") == 0) && (std::strcmp(attr[i + 1], "yes") == 0)) {
				ignore_always = true;
			}
		}
	} else if (std::strcmp(el, "allow") == 0) {
		parseAssert(state == POLICY);
		state = RULE;
		parseRule(Decision::ALLOW, attr);
	} else if (std::strcmp(el, "deny") == 0) {
		parseAssert(state == POLICY);
		state = RULE;
		parseRule(Decision::DENY, attr);
	} else if (std::strcmp(el, "check") == 0) {
		parseAssert(state == POLICY);
		state = RULE;
		parseRule(Decision::CHECK, attr);
	}
}

void XmlParser::elementEnd(const char *el) {
	if (std::strcmp(el, "busconfig") == 0) {
		parseAssert(state == BUSCONFIG);
		state = IDLE;
	} else if (std::strcmp(el, "includedir") == 0) {
		parseAssert(state == INCLUDEDIR);
		state = BUSCONFIG;
		getIncludedFiles(curr_dir, current_text, included_files);
	} else if (std::strcmp(el, "include") == 0) {
		parseAssert(state == INCLUDE);
		state = BUSCONFIG;
		IncludeItem item = parseIncludeItem();
		if (item.ignore_always) {
			return;
		}
		if (access(item.filename.c_str(), F_OK) != 0) {
			if (item.ignore_missing) {
				return;
			}
			tslog::logError(std::string("Missing required policy file: ").append(item.filename).c_str());
			ret_code = -1;
			return;
		}
		included_files.push_back(item.filename);
	} else if (std::strcmp(el, "policy") == 0) {
		parseAssert(state == POLICY);
		state = BUSCONFIG;
	} else if (std::strcmp(el, "allow") == 0) {
		parseAssert(state == RULE);
		state = POLICY;
	} else if (std::strcmp(el, "deny") == 0) {
		parseAssert(state == RULE);
		state = POLICY;
	} else if (std::strcmp(el, "check") == 0) {
		parseAssert(state == RULE);
		state = POLICY;
	}
}

int XmlParser::parsePolicyConfigFile() {
	tslog::log_debug("XmlParser::parsePolicyConfigFile called with filename: ", main_filename);
	parsePolicyConfigFileInternal(main_filename);
	return ret_code;
}

void XmlParser::parsePolicyConfigFileInternal(const char *filename) {
	ret_code = 0;
	try {
		parseXmlFile(filename);
	} catch (const std::runtime_error& ex) {
		tslog::logError(std::string(filename).append(": Error parsing xml: Exception occured: ").append(ex.what()).c_str());
		if (isMainConfFile(filename)) {
			tslog::logError("Exiting parsing");
			ret_code = -1;
			return;
		}
	} catch (...) {
		tslog::logError(std::string(filename).append(" :Error parsing xml: Exception occured: "));
		if (isMainConfFile(filename)) {
			tslog::logError("Exiting parsing");
			ret_code = -1;
			return;
		}
	}
	std::vector<std::string> curr_included_files = included_files;	// deep copy
	for (const auto& included_file : curr_included_files) {
		parsePolicyConfigFileInternal(included_file.c_str());
	}
}

std::unique_ptr<char[]> file2str(const char *filename) {

	FILE *fp = fopen(filename, "rb");
	if (fp == nullptr) {
		throw std::runtime_error(std::string("Failed to open file: ").append(filename).c_str());
	}

	if (fseek(fp, 0, SEEK_END)) {
		throw std::runtime_error(std::string("Failed to fseek end of file").c_str());
	}
	long fsize = ftell(fp);
	if (fsize < 0) {
		throw std::runtime_error(std::string("Failed to ftell").c_str());
	}
	if (fsize > (MAX_CONF_SIZE)) {
		throw std::runtime_error(std::string("File size over ").append(std::to_string(MAX_CONF_SIZE)).append("Bytes").c_str());
	}
	if (fseek(fp, 0, SEEK_SET)) {
		throw std::runtime_error(std::string("Failed to fseek beginning of file").c_str());
	}

	auto str = std::unique_ptr<char[]>(new char[fsize + 1]);	// TODO change to make_unique when c++14 becomes available
	const int count = 1;
	size_t res = fread(str.get(), fsize, count, fp);
	parseAssert(res == count);
	fclose(fp);
	str.get()[fsize] = 0;	// TODO change to operator[] when c++17 becomes available

	return str;
}

void XmlParser::parseXmlFile(const char *filename) {
	included_files.clear();

	tslog::log_debug("Processing: ", filename, " ...\n");
	tslog::log_verbose("=== XML PARSING BEGIN === : ", filename, '\n');

	curr_dir = getDir(filename);

	auto parser = std::unique_ptr<XML_ParserStruct, std::function<void(XML_Parser)>>(XML_ParserCreate(nullptr),
		[](XML_Parser p) { XML_ParserFree(p); });
	XML_SetUserData(parser.get(), this);
	XML_SetElementHandler(parser.get(), start_element_handler, end_element_handler);
	XML_SetCharacterDataHandler(parser.get(), text_handler);

	auto str = file2str(filename);
	if (XML_Parse(parser.get(), str.get(), strlen(str.get()), XML_TRUE) == XML_STATUS_ERROR) {
		throw std::runtime_error(std::string("Error parsing xml file: ").append(XML_ErrorString(XML_GetErrorCode(parser.get()))).c_str());
	}

	tslog::log_verbose("=== XML PARSING END ===\n\n");
}

XmlParser::IncludeItem XmlParser::parseIncludeItem() {
	IncludeItem item;
	item.filename = expandPath(curr_dir, current_text);
	item.ignore_missing = ignore_missing;
	item.ignore_always = ignore_always;
	return item;
}

void XmlParser::getIncludedFiles(const std::string& parent_dir, const std::string& incldir, std::vector<std::string>& files) {
	DIR *dir;
	struct dirent *ent;
	const std::string dname = expandPath(parent_dir, incldir);
	files.clear();
	if ((dir = opendir(dname.c_str())) != NULL) {
		while ((ent = readdir(dir)) != NULL) {
			std::string s(ent->d_name);
			if (boost::algorithm::ends_with(s, ".conf")) {
				files.push_back(dname + std::string("/") + s);
			}
		}
		closedir(dir);
		sort(files.begin(), files.end());

		if (tslog::enabled()) {
			std::cout << "\nincludedir: " << incldir << ", " << files.size() << " included files found:\n";
			std::copy(files.begin(), files.end(), std::ostream_iterator<std::string>(std::cout, "\n"));
			std::cout << '\n';
		}
	} else {
		tslog::log_error("could not open directory ", dname, '\n');
	}
}

void XmlParser::parsePolicyTag(const char **attr)
{
	policy_type = PolicyType::NONE;

	for (int i = 0; attr[i]; i += 2) {
		parsePolicyAttribute(attr[i], attr[i + 1]);
	}
}

void XmlParser::parsePolicyAttribute(const char* name, const char* value)
{
	//possible values (from dbus-daemon man):
	//	context="(default|mandatory)"
	//	at_console="(true|false)" (deprecated and not used on targets, we ignore it)
	//	user="username or userid"
	//	group="group name or gid"
	if (std::strcmp(name, "context") == 0) {
		policy_type = PolicyType::CONTEXT;
		if (std::strcmp(value, "mandatory") == 0 ) {
			policy_type_value.context = ContextType::MANDATORY;
		} else if (std::strcmp(value, "default") == 0) {
			policy_type_value.context = ContextType::DEFAULT;
		}
	} else if (std::strcmp(name, "user") == 0) {
		policy_type = PolicyType::USER;
		policy_type_value.user = convertToUid(value);
	} else if (std::strcmp(name, "group") == 0) {
		policy_type = PolicyType::GROUP;
		policy_type_value.group = convertToGid(value);
	}
}

void XmlParser::parseRule(Decision decision, const char **attr)
{
	if (policy_type == PolicyType::NONE) {
		return;
	}

	__builder.addDecision(decision);

	for (int i = 0; attr[i]; i += 2) {
		parseRuleAttribute(attr[i], attr[i + 1]);
	}

	__builder.generateItem(policy_type, policy_type_value);
}

namespace {
MessageType __str_to_message_type(const char* str) {
	if (!str)
		return MessageType::ANY;

	if (!std::strcmp(str, "method_call"))
		return MessageType::METHOD_CALL;
	else if (!std::strcmp(str, "method_return"))
		return MessageType::METHOD_RETURN;
	else if (!std::strcmp(str, "error"))
		return MessageType::ERROR;
	else if (!std::strcmp(str, "signal"))
		return MessageType::SIGNAL;

	return MessageType::ANY;
}
}

void XmlParser::parseRuleAttribute(const char *name, const char *value)
{
	auto field_has = [] (const std::string& str, const std::string& substr) {
		return (str.find(substr) != std::string::npos);
	};

	if (std::strcmp(value, "*") == 0)
		value = nullptr;

	if (field_has(name, "send_")) {
		__builder.setSendItem();
	} else if (field_has(name, "receive_")) {
		__builder.setReceiveItem();
	} else if (std::strcmp(name, "own") == 0) {
		__builder.addOwner(value);
		__builder.setPrefix(false);
	} else if (std::strcmp(name, "own_prefix") == 0) {
		__builder.addOwner(value);
		__builder.setPrefix(true);
	} else if (std::strcmp(name, "privilege") == 0) {
		__builder.addPrivilege(value);
	} else if (std::strcmp(name, "user") == 0) {
		__builder.setAccessItem();
		if (value)
			__builder.setUserAccess(convertToUid(value));
		else
			__builder.setAllUsersAccess();
	} else if (std::strcmp(name, "group") == 0) {
		__builder.setAccessItem();
		if (value)
			__builder.setGroupAccess(convertToGid(value));
		else
			__builder.setAllGroupsAccess();
	}

	if (field_has(name, "send_destination_prefix"))
		__builder.addName(value, true);
	else if (field_has(name, "_destination"))
		__builder.addName(value);
	else if (field_has(name, "_sender"))
		__builder.addName(value);
	else if (field_has(name, "_path"))
		__builder.addPath(value);
	else if (field_has(name, "_interface"))
		__builder.addInterface(value);
	else if (field_has(name, "_member"))
		__builder.addMember(value);
	else if (field_has(name, "_type"))
		__builder.addMessageType(__str_to_message_type(value));
}

XmlParser::XmlParser(ldp_xml::StorageBackendXML &storage, const char *fname)
	: __builder{storage}, main_filename{fname}
{}
