#include "internal/include/fb_generated.h"
#include "internal/naive_policy_checker.hpp"
#include "internal/policy.hpp"
#include "internal/serializer.hpp"
#include "internal/storage_backend_serialized.hpp"
#include "internal/storage_backend_xml.hpp"
#include "internal/tslog.hpp"
#include <map>

using namespace ldp_xml_parser;
using namespace ldp_serialized;

std::map<Decision, const char*> DECISIONS {
		{Decision::ANY,   "ANY"   },
		{Decision::ALLOW, "ALLOW" },
		{Decision::DENY,  "DENY"  },
		{Decision::CHECK, "CHECK" }
};

enum MessageDirection {
	RECEIVE,
	SEND
};

struct MethodTest {
	Decision expected_result;
	uid_t user;
	gid_t group;
	const char* label;
	const char* name;
	const char* path;
	const char* interface;
	const char* member;
	MessageType type;
	MessageDirection recv_send;
};

/**
 * This test set tests ability to parse xml db
 * and check method call allowance in many use cases
 */

struct MethodTest method_tests[]={
	{Decision::ALLOW, 0,    0,   "test", "org.test.test2",                         NULL, "org.test.Itest1", "DoIt",     MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::ALLOW, 0,    0,   "test", "org.test.test3",                         NULL, "org.test.Itest1", "DoIt",     MessageType::METHOD_CALL, MessageDirection::RECEIVE },

	{Decision::ALLOW, 5001, 100, "test", "org.test.test3",                         NULL, "org.test.Itest1", "DoIt",     MessageType::METHOD_CALL, MessageDirection::RECEIVE },
	{Decision::ALLOW, 0,    0,   "test", "org.test.test2",                         NULL, "org.test.Itest1", "DoIt",     MessageType::METHOD_CALL, MessageDirection::SEND    },

	{Decision::DENY,  0,    0,   "test", "org.test.test2",                         NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::ALLOW, 0,    0,   "test", "org.test.test3",                         NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::RECEIVE },

	{Decision::DENY,  0,    0,   "test", "org.test.test2",                         NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::DENY,  5001, 100, "test", "org.test.test3",                         NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::RECEIVE },

	{Decision::ALLOW, 0,    0,   "test", "test.te34.fg4 a.b.c.d.e org.test.test2", NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::DENY,  0,    0,   "test", "test.te34.fg4 a.b.c.d.e",                NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::ALLOW, 0,    0,   "test", "org.test.test3",                         NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::RECEIVE },

	{Decision::ALLOW, 0,    0,   "test", "org.test.test2",                         NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::DENY,  5001, 100, "test", "org.test.test3",                         NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::RECEIVE },

	{Decision::DENY,  0,    0,   "test", "org.test.test2",                         NULL, "org.test.Itest2", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND    },
	{Decision::ALLOW, 5001, 100, "test", "org.test.test3",                         NULL, "org.test.Itest2", "NotKnown", MessageType::METHOD_CALL, MessageDirection::RECEIVE },
};

void methodTest_print(const struct MethodTest* t, Decision result) {
	printf("uid = %lu, gid = %lu, label = %s, name = %s, path = %s, interface = %s, member = %s, expected = %s, result = %s  (type=%d)",
		   (unsigned long)t->user, (unsigned long)t->group, t->label, t->name, t->path, t->interface, t->member, DECISIONS[t->expected_result], DECISIONS[result], (int)t->recv_send);
}

template <typename DB, typename T>
Decision get_decision(DB &db, const MethodTest &test) {
	DecisionItem ret;

	KdbusBusNames names;
	T m_item = T(test.interface, test.member, test.path, test.type,
			names.addSpaceSeparatedNames(test.name));

	ret = db.getDecisionItemContextMandatory(m_item);
	if (ret.getDecision() == Decision::ANY) {
		ret = db.getDecisionItemUser(test.user, m_item);
	}
	if (ret.getDecision() == Decision::ANY) {
		ret = db.getDecisionItemGroup(test.group, m_item);
	}
	if (ret.getDecision() == Decision::ANY) {
		ret = db.getDecisionItemContextDefault(m_item);
	}
	return ret.getDecision();
}

template <typename DB>
bool method_test(DB &db) {
	unsigned int i = 0;
	bool flag = true;
	Decision decision;

	for (const auto& test : method_tests) {
		if (test.recv_send == MessageDirection::SEND) {
			decision = get_decision<DB, MatchItemSend>(db, test);
		} else {
			decision = get_decision<DB, MatchItemReceive>(db, test);
		}

		bool res = decision == test.expected_result;
		if (!res) {
			printf("[ERROR][%d] method test failed: %s %s ", i++, DECISIONS[test.expected_result],
															   DECISIONS[decision]);
			methodTest_print(&test, decision);
			printf("\n");
			flag = false;
		}
	}
	return flag;
}

bool run_policy_db() {
	policy_checker_system().initDb("tests/default_allow/system.conf");
	auto &db = policy_checker_system().getPolicyDb();
	printf("POLICY_DB:\n");
	return method_test(db);
}

bool run_fb() {
	Serializer serializer;

	size_t size;
	uint8_t *buff = serializer.serialize("tests/default_allow/system.conf", size);
	const FB::File *file = FB::GetFile(buff);

	StorageBackendSerialized storage;
	storage.init(file);
	printf("FLATBUFFERS:\n");
	bool ret = method_test(storage);
	return ret;
}

bool run_xml() {
	// This will be filled in the future
	return true;
}

bool run_tests() {
	return run_policy_db() && run_fb() && run_xml();
}

int main() {
	tslog::init();
	if (!run_tests())
		return -1;
return 0;
}
