#include "internal/naive_policy_checker.hpp"
#include "internal/policy.hpp"
#include "internal/serializer.hpp"
#include "internal/storage_backend_serialized.hpp"
#include "internal/storage_backend_serialized.hpp"
#include "internal/tslog.hpp"
#include <map>

using namespace ldp_xml_parser;
using namespace ldp_serialized;

struct OwnershipTest {
	Decision expected_result;
	uid_t user;
	gid_t group;
	const char* label;
	const char* service;
};

std::map<Decision, const char*> DECISIONS {
	{ Decision::ANY,   "ANY"   },
	{ Decision::ALLOW, "ALLOW" },
	{ Decision::DENY,  "DENY"  },
	{ Decision::CHECK, "CHECK" }
};

/**
 * This test set tests ability to parse xml db
 * and check ownership privilege in many use cases
 * including prefix feature
 */
struct OwnershipTest ownership_tests[]={
	{Decision::ALLOW, 0,    0, "test", "org.test.test1" },
	{Decision::ALLOW, 5009, 0, "test", "org.test.test1" },

	{Decision::DENY,  0,    0, "test", "org.test.test2" },
	{Decision::DENY,  5009, 0, "test", "org.test.test2" },

	{Decision::DENY,  0,    0, "test", "org.test.test3" },
	{Decision::DENY,  5009, 0, "test", "org.test.test3" },

	{Decision::DENY,  0,    0, "test", "org.test.test4" },
	{Decision::ALLOW, 5009, 0, "test", "org.test.test4" },

	{Decision::DENY,  0,    0, "test", "org.test.test5" },

	{Decision::ALLOW, 0,    0, "test", "org.test.test6" },

	{Decision::ALLOW, 0,    0, "test", "org.test.test7" },

	{Decision::ALLOW, 0,    0, "test", "a.b.c"          },
	{Decision::ALLOW, 0,    0, "test", "a.b"            },
	{Decision::DENY,  0,    0, "test", "c"              },
	{Decision::DENY,  0,    0, "test", "a.c"            },
	{Decision::DENY,  0,    0, "test", "b.c"            },

	{Decision::ALLOW, 1000, 0, "test", "org.test.testX" },
	{Decision::ALLOW, 1000, 0, "test", "org.test.testY" }
};

void ownershipTest_print(const struct OwnershipTest* t, const Decision result) {
	printf("uid = %lu, gid = %lu, label = %s, service = %s, expected = %s, result = %s",
		   (unsigned long)t->user, (unsigned long)t->group, t->label, t->service, DECISIONS[t->expected_result], DECISIONS[result]);
}

template <typename DB>
bool ownership_test(const DB &db) {
	unsigned  i = 0;
	bool flag = true;
	for (const auto& test : ownership_tests) {
		auto m_item = MatchItemOwn(test.service);

		auto ret = db.getDecisionItemContextMandatory(m_item);

		if (ret.getDecision() == Decision::ANY)
			ret = db.getDecisionItemUser(test.user, m_item);

		if (ret.getDecision() == Decision::ANY)
			ret = db.getDecisionItemGroup(test.group, m_item);

		if (ret.getDecision() == Decision::ANY)
			ret = db.getDecisionItemContextDefault(m_item);

		auto decision = ret.getDecision();

		if (decision != test.expected_result) {
			printf("[ERROR][%d] ownership test failed: %s %s ", i, DECISIONS[test.expected_result],
				   DECISIONS[decision]);
			ownershipTest_print(&test, decision);
			printf("\n");
			flag = false;
		}
	}
	return flag;
}

bool run_policy_db() {
	policy_checker_system().initDb("tests/default_allow/system.conf");
	auto &db = policy_checker_system().getPolicyDb();
	printf("POLICY DB:\n");
	return ownership_test(db);
}

bool run_fb() {
	Serializer serializer;
	size_t size;
	uint8_t *buff = serializer.serialize("tests/default_allow/system.conf", size);
	const FB::File *file = FB::GetFile(buff);

	StorageBackendSerialized storage;
	storage.init(file);
	printf("FLATBUFFERS:\n");
	return ownership_test(storage);
}

bool run_xml() {
	// This will be filled in the future
	return true;
}

bool run_tests() {
	return run_policy_db() && run_fb() && run_xml();
}

int main() {
	tslog::init();
	if (!run_tests())
		return -1;
	return 0;
}
