#include "internal/naive_policy_checker.hpp"
#include "internal/policy.hpp"
#include "internal/tslog.hpp"

using ldp_xml_parser::MatchItemSend;
using ldp_xml_parser::DecisionResult;

struct Test {
	DecisionResult expected_result;
	uid_t user;
	gid_t group;
	const char *label;
	const char *destination;
	const char *path;
	const char *interface;
	const char *member;
	ldp_xml_parser::MessageType type;
};

const int ROOT = 0;

#define TC(expected_result, names) \
	{(expected_result), ROOT, ROOT, "User::Shell", (names), "/", "a.b", "d", ldp_xml_parser::MessageType::METHOD_CALL}

/**
 * This test set tests ability to parse xml db
 * and check sending privilege in use cases
 * checking send_destination_prefix
 */
struct Test tests[]={
	/* straight-forward tests - base allow */
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.f.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.apf"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.apf.f.f.f.f"),
	/* multiple names owned */
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ao org.tizen.test.dest_prefix.ap.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.f org.tizen.test.dest_prefix.ao"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.do org.tizen.test.dest_prefix.ap.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.f org.tizen.test.dest_prefix.do"),
	/* target holes in default allow */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.1.d"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.1.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.1.dp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.1.dp.f.f.f.f org.tizen.test.dest_prefix.ao"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.1.dp.f.f.f.f org.tizen.test.dest_prefix.ap"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ao org.tizen.test.dest_prefix.ap.1.dp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap org.tizen.test.dest_prefix.ap.1.dp.f.f.f.f"),
	/* target holes in holes in default allow */
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.1.d.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.1.d.ap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.1.dp.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.1.dp.ap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.1.dp.a"),
	/* check redefinitions in default allow */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp.dp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp.dp.ap"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp.dp.ap.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ap.2.apxdp.dp.ap.d"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.2.apxdp.dp.a"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.2.apxdp.dp.ap.f.a"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.2.apxdp.f.f.f.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.2.apxdp.f.f.f.ap.f.f.f"),
	/* totally cancelling previous definitions in default allow */
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap.dp"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap.dp.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap.dp.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap.dp.ap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ap.3.dpxap.ap.dp.a"),
	/* straight-forward tests - base deny */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.f.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dpf"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dpf.f.f.f.f"),
	/* multiple names owned */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.do org.tizen.test.dest_prefix.dp.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.f org.tizen.test.dest_prefix.do"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ao org.tizen.test.dest_prefix.dp.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.f org.tizen.test.dest_prefix.ao"),
	/* target holes in default deny */
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.1.a"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.1.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.1.ap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.1.ap.f.f.f.f org.tizen.test.dest_prefix.do"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.1.ap.f.f.f.f org.tizen.test.dest_prefix.dp"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.do org.tizen.test.dest_prefix.dp.1.ap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp org.tizen.test.dest_prefix.dp.1.ap.f.f.f.f"),
	/* target holes in holes in default demy */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.1.a.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.1.a.dp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.1.ap.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.1.ap.dp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.1.ap.d"),
	/* check redefinitions in default deny */
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap.ap"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap.ap.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap.ap.dp"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap.ap.dp.f.f.f.f"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.dp.2.dpxap.ap.dp.a"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.2.dpxap.ap.d"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.2.dpxap.ap.dp.f.d"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.2.dpxap.f.f.f.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.2.dpxap.f.f.f.dp.f.f.f"),
	/* totally cancelling previous definitions in default deny */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp.ap"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp.ap.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp.ap.dp"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp.ap.dp.f.f.f.f"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.dp.3.apxdp.dp.ap.d"),
	/* checking order in multiple names case */
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.ao org.tizen.test.dest_prefix.do"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.ao.ao org.tizen.test.dest_prefix.do"),
	TC(DecisionResult::DENY, "org.tizen.test.dest_prefix.do org.tizen.test.dest_prefix.ao"),
	TC(DecisionResult::ALLOW, "org.tizen.test.dest_prefix.do org.tizen.test.dest_prefix.ao.ao"),
};

std::map<DecisionResult, const char*> DECISIONS {
		{DecisionResult::CYNARA_ERROR, "CYNARA_ERROR"},
		{DecisionResult::DENY, "DENY"},
		{DecisionResult::ALLOW, "ALLOW"}};

void test_print(struct Test* t, DecisionResult result) {
	printf("uid = %lu, gid = %lu, label = %s, destination = %s, expected = %s, result = %s",
		   (unsigned long)t->user, (unsigned long)t->group, t->label,
		   t->destination, DECISIONS[t->expected_result], DECISIONS[result]);
}

bool test()
{
	unsigned  i = 0;
	bool flag = true;

	policy_checker_system().initDb("tests/default_deny/system.conf");

	for (i = 0; i < sizeof(tests)/sizeof(struct Test); i++) {
		KdbusBusNames names;
		MatchItemSend itemSend(tests[i].interface,
							   tests[i].member,
							   tests[i].path,
							   tests[i].type,
							   names.addSpaceSeparatedNames(tests[i].destination));

		auto decision = policy_checker_system().check(tests[i].user,
													  tests[i].group,
													  tests[i].label,
													  itemSend);
		if (tests[i].expected_result != decision) {
			printf("[ERROR][%d] test failed: %s %s ",
					i, DECISIONS[tests[i].expected_result], DECISIONS[decision]);
			test_print(&tests[i], decision);
			printf("\n");
			flag = false;
		}
	}
	return flag;
}

int main()
{
	tslog::init();
	if (!test())
		return -1;
	return 0;
}
