/* This file contains test-runner for libdbuspolicy
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 * Author: Kazimierz Krosman <k.krosman@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <ctype.h>
#include <time.h>
#include <sys/select.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#include "test_runner.h"

enum {
	PIPE_READ,
	PIPE_WRITE,
};

/*****************CUSTOM SECTION***********************************/
/*custom function that parse stdout of binary and result of binary*/
void parse_test_state(const struct binary* b, const char* test_name, char* buffer, int state_change, int state_option);
void parse_one_test_one_binary(const struct binary* b, const char* test_name, char* buffer, int state_change, int state_option);
char** prepare_args_for_binary(const struct binary* b, const char* test_name);
int daemon_init(void);
int daemon_clear(void);
int cynara_init(void);
int cynara_clear(void);
int cynara_n_init(void);
int cynara_n_clear(void);

static struct test_case ownership_test_cases[] = {
	{"0", "Simple check for ownership in default context"},
	{"1", "Simple check for ownership in default context"},
	{"2", "Ownership check in user context"},
	{"3", "Ownership check in user context (negative test)"},
	{"4", "Ownership check in mandatory context with user context (context priority check)"},
	{"5", "Ownership check in mandatory context with user context (for another user)"},
	{"6", "Ownership check for not root user context (negative test)"},
	{"7", "Ownership check for not root user context"},
	{"8", "Ownership check for group context"},
	{"9", "Ownership check for all context (priority check)"},
	{"10", "Ownership check for all context (priority check)"},
	{"11", "Ownership check for own_prefix case (1)"},
	{"12", "Ownership check for own_prefix case (2)"},
	{"13", "Ownership check for own_prefix case (negative test)"},
	{"14", "Ownership check for own_prefix case (negative test)"},
	{"15", "Ownership check for own_prefix case (negative test)"},
	{NULL, NULL}
};

static struct test_case method_call_test_cases[] = {
	{"0", "check_out test"},
	{"1", "check_in test"},
	{"2", "check_out test"},
	{"3", "check_in test"},
	{"4", "check_out test"},
	{"5", "check_in test"},
	{"6", "check_out test"},
	{"7", "check_in test"},
	{"8", "check_out test"},
	{"9", "check_in test"},
	{"10", "check_out test"},
	{"11", "check_in test"},
	{"12", "check_out test"},
	{"13", "check_in test"},
	{"14", "check_out test"},
	{"15", "check_in test"},
	{"16", "check_out test"},
	{"17", "check_in test"},
	{"18", "check_out test"},
	{"19", "check_in test"},
	{NULL, NULL}
};

static struct test_case signal_test_cases[] = {
	{"0", "Check out for signal"},
	{"1", "Check out for signal (negative)"},
	{NULL, NULL}
};

static struct test_case cynara_test_cases[] = {
	{"0", "Check cynara for result"},
	{NULL, NULL}
};

static struct test_case negative_cynara_test_cases[] = {
	{"0", "Check cynara for result (negative test)"},
	{NULL, NULL}
};

/* This table is used to start binaries */
struct binary tests[] = {
/*path, name, TC_table, timeout in us, prepare_args_handler, parse_function_handler, init_handler, clean_handler*/
	{TEST_PATH "stest_ownership", "ownership", ownership_test_cases,  1000*1000, prepare_args_for_binary, parse_test_state, daemon_init, daemon_clear},
	{TEST_PATH "stest_method_call", "method_call", method_call_test_cases,  1000*1000, prepare_args_for_binary, parse_test_state, daemon_init, daemon_clear},
	{TEST_PATH "stest_signal", "signal", signal_test_cases,  1000*1000, prepare_args_for_binary, parse_test_state, daemon_init, daemon_clear},
	{TEST_PATH "stest_cynara", "cynara", cynara_test_cases,  1000*1000, prepare_args_for_binary, parse_test_state, cynara_init, cynara_clear},
	{TEST_PATH "stest_cynara", "negative_cynara", negative_cynara_test_cases,  1000*1000, prepare_args_for_binary, parse_test_state, cynara_n_init, cynara_n_clear},
};


static const char result_pattern[] = "[r][%0]%1";
static struct state {
	unsigned int i; // index of input buffer
	unsigned int j; // index in results[n]
	unsigned int n; //index in results buffer (0 or 1)
	char results[2][MAX_COMMENT];
} g_state;

static void sm_reset(void)
{
	memset(&g_state, 0, sizeof(g_state));
}

static int sm_update(char* buffer, int i)
{
	int l = strlen(buffer) + 1;
	while (i < l  && g_state.i < sizeof(result_pattern) - 1) {
		if (result_pattern[g_state.i] == '%') {
			g_state.n = result_pattern[g_state.i+1] - '0';
			if (g_state.n > 1) {
				sm_reset();
				i--;
			} else if (isalnum(buffer[i])) {
				g_state.results[g_state.n][g_state.j++] = buffer[i];
			} else if (buffer[i] == result_pattern[g_state.i+2] || buffer[i] == '\n') {
				g_state.results[g_state.n][g_state.j] = 0;
				g_state.i += 3;
				g_state.j = 0;
				if (g_state.n == 1)
					return i;
			} else {
				g_state.i = 0;
				g_state.j = 0;
			}
		} else if (result_pattern[g_state.i] == buffer[i]) {
			g_state.i++;
		} else {
			sm_reset();
		}
		i++;
	}

	if (g_state.i >= sizeof(result_pattern) - 1) {
		g_state.results[g_state.n][g_state.j] = 0;
		g_state.i += 3;
		g_state.j = 0;
		if (g_state.n == 1)
			return i;
	}

	return 0;
}

static const char* sm_get_result(int i)
{
	return g_state.results[i];
}

static pid_t daemon_pids[4] = {0};
static int daemons_pipe[4][2];
static char* daemon_argv1[] = {TEST_PATH "dbus_daemon", "-q", NULL, "-n", "org.test.test3", NULL};
static char* daemon_argv2[] = {TEST_PATH "dbus_daemon", "-q", NULL, "-n", "org.test.test2", "-g", "100", "-u", "5001", NULL};
static char* daemon_argv3[] = {TEST_PATH "dbus_daemon", "-q", NULL, "-n", "org.test.test12", "org.test.test9", NULL};
static char* daemon_argv4[] = {TEST_PATH "dbus_daemon", "-q", NULL, "-n", "org.test.test13", "org.test.test10", "org.test.test11", "-g", "100", "-u", "5001", NULL};
static char** daemons[] = { daemon_argv1, daemon_argv2, daemon_argv3, daemon_argv4};

int daemon_init(void)
{
	int i = 0;
	int ready_proc = 0;
	struct timeval tv;
	fd_set rfds;
	int nfds = -1;
	int proc_num = sizeof(daemons)/sizeof(daemons[0]);

	for (; i < proc_num; i++) {
		if (pipe(daemons_pipe[i]) < 0) {
			perror("cannot create pipe");
			goto error1;
		}

		daemon_pids[i] = fork();
		if (!daemon_pids[i]) {
			char fdd[15];
			snprintf(fdd, sizeof(fdd)-1, "%d", daemons_pipe[i][PIPE_WRITE]);
			daemons[i][2] = fdd;
			close(daemons_pipe[i][PIPE_READ]);
			(void)execv(daemons[i][0], daemons[i]);
			exit(0);
		} else if (daemon_pids[i] > 0) {
			close(daemons_pipe[i][PIPE_WRITE]);
		} else
			goto error1;
	}

	tv.tv_sec = 1;
	tv.tv_usec = 0;
	while (1) {
		FD_ZERO(&rfds);
		for (i = 0; i < proc_num; i++) {
			if (daemons_pipe[i][PIPE_READ] > -1)
				FD_SET(daemons_pipe[i][PIPE_READ], &rfds);
		}

		if (ready_proc >= proc_num)
			break;

		nfds = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
		if (nfds > 0) {
			for (i = 0; i < proc_num; i++) {
				if (daemons_pipe[i][PIPE_READ] > -1 && FD_ISSET(daemons_pipe[i][PIPE_READ], &rfds)) {
					++ready_proc;
					close(daemons_pipe[i][PIPE_READ]);
					daemons_pipe[i][PIPE_READ] = -1;
				}
			}
		} else
			goto error1;
	}

	return 1;
error1:
	daemon_clear();
	return 0;
}

int daemon_clear(void)
{
	unsigned int i = 0;
	for (; i < sizeof(daemons)/sizeof(daemons[0]); i++) {
		if (daemon_pids[i])
			kill(daemon_pids[i], SIGKILL);
	}
	return 0;
}

static int call_script(char* path, char* argv[])
{
	int status;
	int r;
	pid_t pid = fork();

	if (pid < 0)
		return 0;
	else if (!pid) {
		(void)execv(path, argv);
		exit(-1);
	}

	r = waitpid(pid, &status, 0);
	if (r > 0) {
		if (WIFEXITED(status))
			return !(WEXITSTATUS(status));
		else if (WIFCONTINUED(status))
			kill(pid, SIGKILL);
	}

	/* fail */
	return 0;
}

int cynara_n_init(void)
{
	char* argv[] = {CYNARA_PATH, "DENY", NULL};
	return call_script(argv[0], argv);
}

int cynara_n_clear(void)
{
	return cynara_clear();
}

int cynara_init(void)
{
	char* argv[] = {CYNARA_PATH, "ALLOW", NULL};
	return call_script(argv[0], argv);
}

int cynara_clear(void)
{
	char* argv[] = {CYNARA_PATH, NULL};
	return call_script(argv[0], argv);
}

static char* args[3];
char** prepare_args_for_binary(const struct binary* b, const char* test_name)
{
	args[0] = (char*)b->name;
	if (!test_name[0])
		args[1] = NULL;
	else {
		args[1] = (char*)test_name;
		args[2] = NULL;
	}
	return args;
}


/*
 * Example of test_result detector function implementation
 */
void parse_test_state(const struct binary* b, const char* test_name, char* buffer, int state_change, int state_option)
{
	char test_id[MAX_COMMENT];

	switch (state_change) {
	case INIT_TEST:
		sm_reset();
		break;
	case NEW_STDOUT:
		{
			int k = 0;
			buffer[state_option] = 0;
			get_test_id(test_id, sizeof test_id, b, test_name);
			fprintf(stderr, "[stdout][%s]%s\n", test_id, buffer);
			while ((k = sm_update(buffer, k))) {
				const char* t_name = sm_get_result(0);
				const char* result_code = sm_get_result(1);
				bool is_successful = result_code[0] == '1' && result_code[1] == 0;
				const char* result_text = is_successful ? "PASS" : "FAIL";

				get_test_id(test_id, sizeof test_id, b, t_name);
				add_test_result(test_id, result_text, "", is_successful);
				sm_reset();
			}
		}
		break;
	case NEW_STDERR:
		buffer[state_option] = 0;
		get_test_id(test_id, sizeof test_id, b, test_name);
		fprintf(stderr, "[stderr][%s]%s\n", test_id, buffer);
		break;
	case RESULT_CODE:
		if (state_option != 0)
			add_test_result(get_test_id(test_id, sizeof test_id, b, test_name), "FAIL", "Test exited with error code", 0);
		break;
	case RESULT_SIGNAL:
		add_test_result(get_test_id(test_id, sizeof test_id, b, test_name), "FAIL", "Finished by SIGNAL", 0);
		break;
	case RESULT_TIMEOUT:
		add_test_result(get_test_id(test_id, sizeof test_id, b, test_name), "FAIL", "Test TIMEOUT", 0);
		break;
	}
}

/****************************END of CUSTOM FUNCTIONS************************/


static struct option long_options[] = {
	{"list",        no_argument,       0, 'l'},
	{"run",         required_argument, 0, 'r'},
	{"description", required_argument, 0, 'd'},
	{0,             0,                 0,  0 }
};

static int stdin_pipe[2];
static int stdout_pipe[2];
static int stderr_pipe[2];
static struct test_result test_results[MAX_TC_NUM];
static int test_results_i;
static char buffer[MAX_BUFFER];
static const char* requested_tc[MAX_TC_NUM];

char* get_test_id(char *dest, size_t dest_len, const struct binary* b, const char* test_name)
{
	const size_t name_len = strlen(b->name);
	const size_t test_len = strlen(test_name);

	assert(name_len + test_len < dest_len);

	memcpy(dest, b->name, name_len);
	memcpy(dest + name_len, test_name, test_len + 1);
	return dest;
}

static void print_description(const char* name, const char* description)
{
	printf("%s;%s\n", name, description);
}

static void print_list(const char* test_name)
{
	unsigned int i;
	char full_name[MAX_COMMENT];
	for (i = 0; i < sizeof(tests)/sizeof(struct binary); i++) {
		int j = 0;
		int l = strlen(tests[i].name);
		memcpy(full_name, tests[i].name, l+1);
		if (test_name && strncmp(test_name, full_name, l) != 0)
			continue;

		while (tests[i].test_cases[j].name) {
			memcpy(full_name + l, tests[i].test_cases[j].name, strlen(tests[i].test_cases[j].name) + 1);
			if (!test_name || strcmp(full_name, test_name) == 0)
				print_description(full_name, tests[i].test_cases[j].description);
			j++;
		}
	}
}


static void stop_binary(const struct binary* b, pid_t pid, const char* test_name)
{
	int status = 0;
	int res = 0;
	res = waitpid(pid, &status, 0);
	if (res == 0) {
		//timeouted
		kill(pid, SIGKILL);
		res = waitpid(pid, &status, WNOHANG);
		b->parse(b, test_name, buffer, RESULT_TIMEOUT, res);
	} else if (res < 0) {
		//errno check
		kill(pid, SIGKILL);
		res = waitpid(pid, &status, WNOHANG);
		b->parse(b, test_name, buffer, RESULT_ERROR, res);
	} else if (res > 0) {
		if (WIFEXITED(status)) {
			b->parse(b, test_name, buffer, RESULT_CODE, WEXITSTATUS(status));
		} else if (WIFSIGNALED(status)) {
			b->parse(b, test_name, buffer, RESULT_SIGNAL, WTERMSIG(status));
		} else if (WIFSTOPPED(status)) {
			b->parse(b, test_name, buffer, RESULT_SIGNAL, WSTOPSIG(status));
		} else if (WIFCONTINUED(status)) {
			kill(pid, SIGKILL);
			b->parse(b, test_name, buffer, RESULT_SIGNAL, -1);
		}
	}
}

static void parse_output_with_timeout(const struct binary* b, pid_t pid, const char* test_name)
{
	struct timeval tv;
	fd_set rfds;
	int nfds;
	int res;
	tv.tv_sec = b->timeout / (1000*1000);
	tv.tv_usec = (b->timeout-tv.tv_sec * 1000*1000);
	while (1) {
		FD_ZERO(&rfds);
		FD_SET(stdout_pipe[PIPE_READ], &rfds);
		FD_SET(stderr_pipe[PIPE_READ], &rfds);
		nfds = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);
		if (nfds == -1) {
			if (errno != EINTR)
				break;
		} else if (nfds > 0) {
			if (FD_ISSET(stdout_pipe[PIPE_READ], &rfds)) {
				res = read(stdout_pipe[PIPE_READ], buffer, MAX_BUFFER-1);
				if (res == 0 || (res < 0 && errno != EINTR))
					break;
				else if (res >= 0)
					b->parse(b, test_name, buffer, NEW_STDOUT, res);
			}

			if (FD_ISSET(stderr_pipe[PIPE_READ], &rfds)) {
				res = read(stderr_pipe[PIPE_READ], buffer, MAX_BUFFER-1);
				if (res == 0 || (res < 0 && errno != EINTR))
					break;
				b->parse(b, test_name, buffer, NEW_STDERR, res);
			}
		} else {
			//timeout
			break;
		}
	}
	stop_binary(b, pid, test_name);
}

static int create_child(const char* path, char* const arguments[])
{
	int child;
	int nResult;
	if (pipe(stdin_pipe) < 0) {
		perror("allocating pipe for child input redirect failed");
		goto error1;
	}

	if (pipe(stdout_pipe) < 0) {
		perror("allocating pipe for child output redirect failed");
		goto error2;
	}

	if (pipe(stderr_pipe) < 0) {
		perror("allocating pipe for child output redirect failed");
		goto error3;
	}

	child = fork();
	if (!child) {
		char ld_path[512];
		snprintf(ld_path, sizeof(ld_path)-1, "/usr/lib/dbus-tests/lib/libdbuspolicy-tests/:");
		// redirect stdin
		if (dup2(stdin_pipe[PIPE_READ], STDIN_FILENO) == -1) {
			perror("redirecting stdin failed");
			return -1;
		}

		// redirect stdout
		if (dup2(stdout_pipe[PIPE_WRITE], STDOUT_FILENO) == -1) {
			perror("redirecting stdout failed");
			return -1;
		}

		// redirect stderr
		if (dup2(stderr_pipe[PIPE_WRITE], STDERR_FILENO) == -1) {
			perror("redirecting stderr failed");
			return -1;
		}

		// all these are for use by parent only
		close(stdin_pipe[PIPE_READ]);
		close(stdin_pipe[PIPE_WRITE]);
		close(stdout_pipe[PIPE_READ]);
		close(stdout_pipe[PIPE_WRITE]);
		close(stderr_pipe[PIPE_READ]);
		close(stderr_pipe[PIPE_WRITE]);

		char* ld_path_b = getenv("LD_LIBRARY_PATH");
		if (ld_path_b != NULL)
			memcpy(ld_path + strlen(ld_path), ld_path_b, strlen(ld_path_b)+1);
		setenv("LD_LIBRARY_PATH", ld_path, 1);
		// run child process image
		nResult = execv(path, arguments);

		// if we get here at all, an error occurred, but we are in the child
		// process, so just exit
		perror("exec of the child process  failed");
		exit(nResult);
	} else if (child > 0) {
		// parent continues here

		// close unused file descriptors, these are for child only
		close(stdin_pipe[PIPE_READ]);
		close(stdout_pipe[PIPE_WRITE]);
		close(stderr_pipe[PIPE_WRITE]);
	} else {
		// failed to create child
		goto error4;
	}

	return child;

error4:
	close(stderr_pipe[PIPE_READ]);
	close(stderr_pipe[PIPE_WRITE]);
error3:
	close(stdout_pipe[PIPE_READ]);
	close(stdout_pipe[PIPE_WRITE]);
error2:
	close(stdin_pipe[PIPE_READ]);
	close(stdin_pipe[PIPE_WRITE]);
error1:
	return -1;
}

static void run_test(const struct binary* b, const char* test_name)
{
	int res = -1;
	char** arg;
	char test_id[MAX_COMMENT];

	assert(b);
	assert(b->name);
	assert(b->path);
	assert(test_name);

	arg = b->prepare_args(b, test_name);

	if (b->init)
		if (!b->init()) {
			add_test_result(get_test_id(test_id, sizeof test_id, b, test_name), "ERROR", "Internal error: Cannot init test", 0);
			return;
		}

	res = create_child(b->path, arg);
	if (res > 0)
		parse_output_with_timeout(b, res, test_name);
	else
		add_test_result(get_test_id(test_id, sizeof test_id, b, test_name), "ERROR", "Internal error: Cannot start test", 0);

	if (b->clean)
		b->clean();
}

static void parse_run_test(const char* tc)
{
	unsigned int i = 0;
	for (i = 0; i < sizeof(tests)/sizeof(struct binary); i++) {
		int len = strlen(tests[i].name);
		if (strncmp(tc, tests[i].name, len) == 0) {
			if (tc[len] == '*' || tc[len] == '\0')
				run_test(&tests[i], "");
			else
				run_test(&tests[i], tc + len);
		}
	}
}

static int parse_option(int argc, char* argv[])
{
	int ch = 0;
	int c = 0;
	while ((ch = getopt_long(argc, argv, "lr:d:", long_options, NULL)) != -1) {
		switch (ch) {
		case 'l':
			print_list(NULL);
			return 1;
		case 'r':
			if (c >= MAX_TC_NUM - 1) //NULL at the end
				return 0;

			if (optarg)
				requested_tc[c++] = optarg;

			break;
		case 'd':
			print_list(optarg);
			return 1;
		}
	}
	return 0;
}

//output_new/fail/success
void add_test_result(const char* test_id, const char* result, const char* comment, int res)
{
	test_results[test_results_i].is_positive = res;
	strncpy(test_results[test_results_i].result, result, sizeof(test_results[test_results_i].result) - 1);
	strncpy(test_results[test_results_i].comment, comment, sizeof(test_results[test_results_i].comment) - 1);
	strncpy(test_results[test_results_i].name, test_id, sizeof(test_results[test_results_i].name) - 1);
	test_results_i++;
}

static void prepare_results(void)
{
	//todo
}

static void print_results()
{
	int i = 0;
	for (i = 0; i < test_results_i; i++)
		printf("%s;%s;%s\n", test_results[i].name, test_results[i].result, test_results[i].comment);
}

int main(int argc, char* argv[])
{
	unsigned int i;
	signal(SIGPIPE, SIG_IGN);
	if (parse_option(argc, argv))
		return 0;

	prepare_results();

	if (!requested_tc[0]) {
		for (i = 0; i < sizeof(tests) / sizeof(struct binary); i++)
		    run_test(&tests[i], "");
	} else {
		i = 0;
		while (requested_tc[i]) {
		    parse_run_test(requested_tc[i]);
			i++;
		}
	}

	print_results();
	return 0;
}
